/**
 * @file fedkit-file.c (in driver section)
 * kernel side driver of the fedkit receiver driver 
 * the kernel side drivers has the following functions
 * at installation time, it has to detect all the receiver boards present
 *
 * first function is to reserve a board (identified by ID) for the calling 
 * file descriptor. Base address for the board is delivered in the reply
 *
 * second call is used to register the address of the data fifo addresses 
 * to the interrupt service routine
 *
 * final one is to release the adapter and to free up everything
 * 
 */ 
/* fedkit documentatio is at http://cern.ch/cano/fedkit/ */
/*
 Maintainer : Eric Cano, CERN-EP/CMD <Eric.Cano@cern.ch>
 
 $Id: fedkit-file.c,v 1.3 2007/10/24 17:36:42 cano Exp $
*/
static char *rcsid = "@(#) $Id: fedkit-file.c,v 1.3 2007/10/24 17:36:42 cano Exp $";
/* The following lines will prevent `gcc' version 2.X
   from issuing an "unused variable" warning. */
#if __GNUC__ >= 2
#define USE(var) void use_##var (void) {var = var;}
USE (rcsid);
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/bigphysarea.h>
#include <asm/uaccess.h>
#include "fedkit-private.h"

/* Driver's local variables */

struct semaphore kfedkit_prev_dbuff_sem; /* for use in fedkit-Dbuff.c */

/* file operations for the fedkit device driver */

/* forward declarations */

int kfedkit_open (struct inode *, struct file *);
int kfedkit_ioctl (struct inode *, struct file *, unsigned int, unsigned long);
int kfedkit_mmap (struct file *, struct vm_area_struct *);
int kfedkit_release (struct inode *, struct file *);

struct file_operations kfedkit_fops = {
        ioctl : kfedkit_ioctl,
        mmap : kfedkit_mmap,
        open : kfedkit_open,
        release : kfedkit_release,
};

struct kfedkit_receiver_t  *kfedkit_receivers = NULL;
struct kfedkit_sender_t  *kfedkit_senders = NULL;
struct semaphore kfedkit_receivers_lock;
struct semaphore kfedkit_senders_lock;

int init_module (void)
{
    printk(KERN_INFO "Loading fedkit version $Id: fedkit-file.c,v 1.3 2007/10/24 17:36:42 cano Exp $\n");
    printk(KERN_INFO "Compiled "__DATE__ " " __TIME__ "\n" );
    init_MUTEX (&kfedkit_receivers_lock)/* = MUTEX*/;
    init_MUTEX (&kfedkit_senders_lock)/* = MUTEX*/;
    init_MUTEX (&kfedkit_prev_dbuff_sem) /* = MUTEX*/;
    if (register_chrdev (FEDKIT_MAJOR, "fedkit", &kfedkit_fops)) {
        printk (KERN_ERR "Failed to register fedkit in init_module\n");
        return -EIO;
    }
    return 0;
}

void cleanup_module (void)
{
    /* Extra safety : collect semaphores */
    down (&kfedkit_receivers_lock);
    down (&kfedkit_senders_lock);
    unregister_chrdev (FEDKIT_MAJOR, "fedkit");
}

/** kfedkit_nail_pages */
void kfedkit_nail_pages (void *kernel_address, int size) 
{
    /* we will nail all the pages */
    struct page * page;
    for (page = virt_to_page(kernel_address); 
         page <= virt_to_page(kernel_address + size -1); 
         page++) 
        /*mem_map_reserve(page); */
        set_bit(PG_reserved, &((page)->flags));
}

/** kfedkit_unnail_pages */
void kfedkit_unnail_pages (void *kernel_address, int size) 
{
    /* we will nail all the pages */
    struct page * page;
    for (page = virt_to_page(kernel_address); 
         page <= virt_to_page(kernel_address + size -1); 
         page++) 
        /*mem_map_unreserve(page); */
        clear_bit(PG_reserved, &((page)->flags));
}

/* register receiver : check of the receiver is allocated, if not,
   look for device, allocate, and fill structures with relevant data */
int kfedkit_register_receiver (int index, struct inode *inode_p, struct file * file_p)
{
    struct kfedkit_receiver_t *receiver;
    struct kfedkit_board_t *board;
    int count;

    idprintf ("In kfedkit_register_receiver\n");
    /* critical section */
    down (&kfedkit_receivers_lock);
    /* find the receiver in the list */
    idprintf ("In kfedkit_register_receiver\n");
    for (receiver = kfedkit_receivers; receiver != NULL; receiver=receiver->next) {
        
        if (receiver->index == index) break;
    }
    idprintf ("In kfedkit_register_receiver\n");
    if (receiver != NULL) {
        /* receiver was already allocated, sorry (EBUSY) */
        up (&kfedkit_receivers_lock);
        return -EBUSY;
    }
    
    idprintf ("In kfedkit_register_receiver\n");
    board = (struct kfedkit_board_t *) kmalloc (sizeof (struct kfedkit_board_t), GFP_KERNEL);
    if (board == NULL) {
        /* could not allocate memory */
        up (&kfedkit_receivers_lock);
        return -ENOMEM;
    }
    idprintf ("In kfedkit_register_receiver\n");
    board->type = receiver_board;
    board->dbuffs = NULL;
    receiver = &board->receiver;
    /* The device is not allocated. Let's find it */
    receiver->device = NULL;
    idprintf ("In kfedkit_register_receiver\n");
    count = 0;
    idprintf ("In kfedkit_register_receiver\n");
    while (NULL != (receiver->device = pci_find_device(_FK_VENDOR, _FK_DEVICE, receiver->device)))
        if (count++ == index)
            break;
    idprintf ("In kfedkit_register_receiver\n");
    if (receiver->device == NULL) {
        /* the required device could not be found */
        kfree (receiver);
        up (&kfedkit_receivers_lock);
        return -ENODEV;
    }
    /* everything seems to be in order. Let's get out of the critical section */
    idprintf ("In kfedkit_register_receiver\n");
    receiver->index = index;
    receiver->next = kfedkit_receivers;
    kfedkit_receivers = receiver;
    /*up (&kfedkit_receivers_lock);*/
    /* now we can fill the blanks in the strucuture, and return a success */
    idprintf ("In kfedkit_register_receiver\n");
    receiver->block_fifo = NULL;
    receiver->blocks = NULL;
    receiver->wc_fifo = NULL;
    receiver->block_number = 0;
    receiver->block_size = 0;
    idprintf ("In kfedkit_register_receiver\n");
    pci_read_config_dword (receiver->device, 0x10, &(receiver->base_address0));
    pci_read_config_dword (receiver->device, 0x14, &(receiver->base_address1));
    pci_read_config_dword (receiver->device, 0x18, &(receiver->base_address2));
    pci_read_config_dword (receiver->device, 0x48, &(receiver->FPGA_version));
    idprintf ("In kfedkit_register_receiver\n");
    /* only base addres 0 need to be mapped in the kernel */
    if (receiver->base_address0!= 0) 
        receiver->map0 = ioremap_nocache (receiver->base_address0, _FK_BAR0_range);
    /*if (receiver->base_address1!= 0) 
      receiver->map1 = ioremap_nocache (receiver->base_address1, _FK_BAR1_range);*/
    receiver->map1 = NULL;
    receiver->map2 = NULL;
    idprintf ("In kfedkit_register_receiver\n");
    if (receiver->map0 == NULL/* || receiver->map1 == NULL*/) {
        /* we could not map in kernel one of the memory zones... Fishy */
        idprintf ("In kfedkit_register_receiver : could not map in kernel one of the base addresses\n");
        idprintf ("BARS : 0x%08x/0x%08x maps : 0x%08x/0x%08x\n", receiver->base_address0, receiver->base_address1,
                  (int)receiver->map0, (int)receiver->map1);
        if (receiver->map0 != NULL)
            iounmap (receiver->map0);
        if (receiver->map1 != NULL)
            iounmap (receiver->map1);
        /* unplug the board from the general list (we still hold the lock) and free structure */
        kfedkit_receivers = receiver->next;        
        kfree (board);
        up (&kfedkit_receivers_lock);
        return -EIO;
    }
    init_MUTEX (&(receiver->lock))/* = MUTEX*/;
    /* keep a handy pointer to the structure */
    file_p->private_data = board;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_INC_USE_COUNT;
#endif
    receiver->interrupt_registered = 0;
    up (&kfedkit_receivers_lock);
    idprintf ("In kfedkit_register_receiver\n");
    return 0;
}

#ifndef FEDKIT_NO_MASTER_INTS


/* This interrupt works in 2 ways, depending on wether or not
   something is available for it to send. If there is something, the
   interrupt is cleared, and the parameters for the new DMA are
   put. If not, the interrupt is disabled, and has to be enabled by
   the user space routine. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
void kfedkit_sender_ISR (int irq, void *sender_p, struct pt_regs* regs)
#else
irqreturn_t kfedkit_sender_ISR (int irq, void *sender_p, struct pt_regs* regs)
#endif
{
	uint32_t csr;
	struct kfedkit_sender_t *sender = (struct kfedkit_sender_t *) sender_p;
	/* check for interrupt */
	
    csr = sender->map0[_FK_SENDER_MCTRL / 4];
    if (csr & _FK_SENDER_MASTER_DMA_DONE) { /* there was an interrupt */
        if (sender->master_fifo->write != sender->master_fifo->read) {
            /* something's available we are safe to pop something from the fifo*/
            int read = sender->master_fifo->read;
            void * address = sender->master_fifo->data[read].address;
            int wc = sender->master_fifo->data[read].word_count;
            sender->master_fifo->inprogress = address;
            sender->master_fifo->read = (read + 1) % sender->master_fifo->size;
            /* Clear the interrupt (there's another DMA to come). Bug
               workaround : sometimes this interrupt comes again after
               the software tried to re-inable it, and the ISR had
               disabled it. They then read both the register in the
               non-enabled state. When we are here, the interrupt
               should be enabled, so we re-enable it (force) */
            sender->map0[_FK_SENDER_MCTRL / 4] = csr | _FK_SENDER_DMA_INT_CLEAR | _FK_SENDER_ENABLE_FIFO_DMA_DONE;
            /* start the new DMA */
            sender->map0[_FK_SENDER_MEVTADDR / 4] = (int)address;
            sender->map0[_FK_SENDER_MEVTSZ / 4] = wc;
        } else {
            /* Let's just disable the interrupt */
            sender->master_fifo->int_disabled = 1;
            sender->map0[_FK_SENDER_MCTRL / 4] = csr & ~_FK_SENDER_ENABLE_FIFO_DMA_DONE;
            sender->master_fifo->inprogress = NULL;
            /* BUG WORKAROUND : the interrupt seems to rise twice
               sometimes because the int signal (open collector)
               doesn't go up fast enough. To work this around, I add a
               read cycle in this int */
            csr = sender->map0[_FK_SENDER_MCTRL / 4];
        }
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        return IRQ_HANDLED;
    } else {
        return IRQ_NONE;
    #endif
    }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)    
    return IRQ_HANDLED; /* just to keep the compiled from bitchin' */
#endif 
}

/** @function kfedkit_register_interrupt
    @param receiver pointer to the receiver
    registers interrupt for this receiver */
int kfedkit_register_sender_interrupt (struct kfedkit_sender_t *sender)
{
    snprintf (sender->device_name, FEDKIT_DEVNAME_SIZE, "fedkit_sender%d", sender->index);
    sender->device_name[FEDKIT_DEVNAME_SIZE -1] = '\0';
    if (request_irq (sender->device->irq, kfedkit_sender_ISR, 
                     SA_SHIRQ,"fedkit_sender"/* sender->device_name*/, sender)) {
        /* irq registering failed */
        iprintf ("INT REG FAILED!\n");
        return -1;
    } else {
        sender->interrupt_registered = -1;
        return 0;
    }
}

void kfedkit_unregister_sender_interrupt (struct kfedkit_sender_t *sender)
{
    if (sender->interrupt_registered)
        free_irq (sender->device->irq, sender);
}
#endif /* ndef FEDKIT_NO_MASTER_INTS */
int kfedkit_register_sender (int index, struct inode *inode_p, struct file * file_p)
{
    struct kfedkit_sender_t *sender;
    struct kfedkit_board_t *board;
    int count;

    /* critical section */
    down (&kfedkit_senders_lock);
    /* find the receiver in the list */
    for (sender = kfedkit_senders; sender != NULL; sender=sender->next) {
        if (sender->index == index) break;
    }
    if (sender != NULL) {
        /* sender was already allocated, sorry (EBUSY) */
        iprintf ("BUSY!\n");
        up (&kfedkit_senders_lock);
        return -EBUSY;
    }
    board = (struct kfedkit_board_t *) kmalloc (sizeof (struct kfedkit_board_t), GFP_KERNEL);
    if (board == NULL) {
        /* could not allocate memory */
        iprintf ("ALLOC!\n");
        up (&kfedkit_senders_lock);
        return -ENOMEM;
    }
    board->type = sender_board;
    board->dbuffs = NULL;
    sender = &board->sender;
#ifndef FEDKIT_NO_MASTER_INTS
    /* allocate the master_FIFO_zone (has to be page aligned because of mmap, so let's use bigphys */
    sender->master_fifo = (struct _fedkit_sender_master_fifo_t *) 
        bigphysarea_alloc_pages ((sizeof (struct _fedkit_sender_master_fifo_t) - 1) / PAGE_SIZE + 1, 0, GFP_KERNEL);
    edprintf ("Allocated bigphys of size 0x%lx, result 0x%x\n", (sizeof (struct _fedkit_sender_master_fifo_t) - 1) / PAGE_SIZE + 1, (unsigned int)sender->master_fifo);
    if (sender->master_fifo == NULL) {
        /* could not allocate memory */
        iprintf ("BIG ALLOC!\n");
        kfree (board);
        up (&kfedkit_senders_lock);
        return -ENOMEM;
        
    }
    kfedkit_nail_pages (sender->master_fifo, sizeof (struct _fedkit_sender_master_fifo_t));
#endif /* ndef FEDKIT_NO_MASTER_INTS */
    /* The device is not allocated. Let's find it */
    sender->device = NULL;
    count = 0;
    while (NULL != (sender->device = pci_find_device(_FK_VENDOR, _FK_SENDER_DEVICE, sender->device)))
        if (count++ == index)
            break;
    if (sender->device == NULL) {
        /* the required device could not be found */
        iprintf ("NODEV!\n");
        up (&kfedkit_senders_lock);
#ifndef FEDKIT_NO_MASTER_INTS
        kfedkit_unnail_pages (sender->master_fifo, sizeof (struct _fedkit_sender_master_fifo_t));
	edprintf ("About to free bigphys 0x%x\n", (unsigned int)sender->master_fifo);
        bigphysarea_free_pages ((caddr_t)sender->master_fifo);
#endif /* ndef FEDKIT_NO_MASTER_INTS */
        kfree (board);
        return -ENODEV;
    }
    /* everything seems to be in order. Let's get out of the critical section */
    sender->index = index;
    sender->next = kfedkit_senders;
    kfedkit_senders = sender;
    up (&kfedkit_senders_lock);
    /* now we can fill the blanks in the strucuture, and return a success */
    pci_read_config_dword (sender->device, 0x10, &(sender->base_address0));
    pci_read_config_dword (sender->device, 0x14, &(sender->base_address1));
    pci_read_config_dword (sender->device, 0x18, &(sender->base_address2));
    pci_read_config_dword (sender->device, 0x48, &(sender->FPGA_version));
#ifndef FEDKIT_NO_MASTER_INTS
    /* We now map the sender's base addresses in kernel space */
    if (sender->base_address0!= 0) 
        sender->map0 = ioremap_nocache (sender->base_address0, _FK_sender_BAR0_range);
    sender->map1 = NULL;
    sender->map2 = NULL;
    if (sender->map0 == NULL) {
        printk (KERN_ERR "FEDKIT: Could not map the sender's BAR0 in kernel space\n");
        kfedkit_unnail_pages (sender->master_fifo, sizeof (struct _fedkit_sender_master_fifo_t));
        edprintf ("About to free bigphys 0x%x\n", (unsigned int)sender->master_fifo);
	bigphysarea_free_pages ((caddr_t)sender->master_fifo);
        kfree (board);
        return -EIO;
    }
    /* We still have to prepare the structure of the fifo (trivial) and to register the interrupt */
    sender->master_fifo->write = 0;
    sender->master_fifo->read = 0;
    sender->master_fifo->size = _FK_MASTER_FIFO_SIZE;
    sender->master_fifo->int_disabled = 1;
    sender->master_fifo->inprogress = NULL;
    sender->master_fifo->first_dma_done = 0;
    sender->interrupt_registered = 0;

    /* now install the interrupt */
    if (kfedkit_register_sender_interrupt(sender)) {
        if (sender->map0 != NULL) {
            iounmap (sender->map0);
        }
        printk (KERN_ERR "FEDKIT: Failed to register the interrupt for the sender\n");
        kfedkit_unnail_pages (sender->master_fifo, sizeof (struct _fedkit_sender_master_fifo_t));
        edprintf ("About to free bigphys 0x%x\n", (unsigned int)sender->master_fifo);
	bigphysarea_free_pages ((caddr_t)sender->master_fifo);
        kfree (board);
        return -EIO;
    }
#endif /* ndef FEDKIT_NO_MASTER_INTS */
    init_MUTEX (&(sender->lock))/* = MUTEX*/;
    /* keep a handy pointer to the structure */
    file_p->private_data = board;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_INC_USE_COUNT;
#endif
    return 0;
}

/* kfedkit_open. Just handles board reservation with kfedkit_close */ 
int kfedkit_open (struct inode *inode_p, struct file * file_p)
{
    /* First let's find out what is required */
    if (MINOR (inode_p->i_rdev) < MAX_FEDKIT_SENDER_NUMBER) {
        /* We look for a sender */
        int sender_index = MINOR (inode_p->i_rdev) ;
        idprintf ("Will call kfedkit_register_sender for sender %d\n", sender_index);
        return kfedkit_register_sender (sender_index, inode_p, file_p);
    } else {
        /* We look for a receiver */
        int receiver_index = MINOR (inode_p->i_rdev) - MAX_FEDKIT_SENDER_NUMBER;
        idprintf ("Will call kfedkit_register_receiver for receiver %d\n", receiver_index);
        return kfedkit_register_receiver (receiver_index, inode_p, file_p);
    }
    return -EFAULT;
}

void kfedkit_unregister_sender (struct kfedkit_board_t *board)
{
    struct kfedkit_sender_t * sender, **pp_sender;
    edprintf ("A\n");
    if (board->type != sender_board) 
        printk (KERN_ERR "Something wrong going on in kfedkit_unregister_sender: not a sender\n");
    sender = &board->sender;
    /* lock */
    edprintf ("A\n");
    down (&sender->lock);
    /* cleanup : soft reset and unregister interrupt */
    edprintf ("A\n");
    if (sender->map0!= NULL) {
        sender->map0[_FK_SENDER_MCTRL / 4] |= _FK_SENDER_SOFT_RESET;
        iounmap (sender->map0);
    }
    edprintf ("A\n");
    /* nothing to do on the sender itself... */
    /* remove the sender from the sender list */
    edprintf ("A\n");
    down (&kfedkit_senders_lock);
    edprintf ("A\n");
    for (pp_sender = &kfedkit_senders; *pp_sender != NULL;) {
        if (*pp_sender == sender) *pp_sender = (*pp_sender)->next;
        else pp_sender = &(*pp_sender)->next;
    }
    edprintf ("A\n");
    up (&kfedkit_senders_lock);
#ifndef FEDKIT_NO_MASTER_INTS
    edprintf ("A\n");
    kfedkit_unregister_sender_interrupt (sender);
    /* free the board structure */
    edprintf ("A\n");
    kfedkit_unnail_pages (sender->master_fifo, sizeof (struct _fedkit_sender_master_fifo_t));
    edprintf ("A\n");
    edprintf ("About to free bigphys 0x%x\n", (unsigned int)sender->master_fifo);
    bigphysarea_free_pages ((caddr_t)sender->master_fifo);
#endif /* ndef FEDKIT_NO_MASTER_INTS */
    edprintf ("A\n");
    kfree (board);
}

void * kfedkit_pop_data (struct kfedkit_receiver_t *receiver)
{
	void *ret;
	int next_read;
	
	idprintf ("In fedkit_pop_data : read = %d, write=%d, ,size=%d, tab=0x%08x\n",
		receiver->data_fifo->read, receiver->data_fifo->write, receiver->data_fifo->size,
		(int)receiver->data_fifo_tab);
	if (receiver->data_fifo->read == receiver->data_fifo->write) {
		return NULL;
	}
	ret = receiver->data_fifo_tab[receiver->data_fifo->read];
	next_read = (receiver->data_fifo->read + 1) % receiver->data_fifo->size;
	receiver->data_fifo->read = next_read;
	receiver->data_fifo->done_read ++;
	return ret;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
void kfedkit_ISR (int irq, void *receiver_p, struct pt_regs* regs)
#else
irqreturn_t kfedkit_ISR (int irq, void *receiver_p, struct pt_regs* regs)
#endif
{
    int target;
    uint32_t csr;
    struct kfedkit_receiver_t *receiver = (struct kfedkit_receiver_t *) receiver_p;
    /* find out which interrupt was triggered (half empty or empty ?) */
    csr = receiver->map0[_FK_CSR_OFFSET / 4];
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    if (!csr & (_FK_CSR_EMPTY_INT_ENABLE | _FK_CSR_HALF_EMPTY_INT_ENABLE)) {
        return IRQ_NONE;
    }
    #endif
    if ((csr & _FK_CSR_EMPTY_INT_STATUS) && (csr & _FK_CSR_EMPTY_INT_ENABLE)) {
        idprintf ("In fedkit interrupt : empty\n");
        target = _FK_free_data_fifo_depth;
    } else if ((csr & _FK_CSR_HALF_EMPTY_INT_STATUS) && (csr & _FK_CSR_HALF_EMPTY_INT_ENABLE)) {
        idprintf ("In fedkit interrupt : half empty\n");
        target = _FK_free_data_fifo_depth / 2;
    } else {
        target = 0;
    }
    idprintf ("In fedkit interrupt, target = %d\n", target);
    for ( ; target > 0; target--) {
        void * data_block = kfedkit_pop_data (receiver);
        if (data_block != NULL) {
            receiver->map0[_FK_FRBKFIFO_OFFSET/4] = (uint32_t)data_block;
        } else {
            target = -1;
            break;
        }
    }
    idprintf ("In fedkit interrupt, target = %d\n", target);
    if (target <= -1) { /* we are unable to feed the FIFO. disable interrupts if needed */
        receiver->map0[_FK_CSR_OFFSET / 4] &= ~_FK_CSR_ALL_INTS_ENABLE;
        receiver->data_fifo->int_disabled = 1;
    }
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    return IRQ_HANDLED;
    #endif
    
}

/** @function kfedkit_register_interrupt
    @param receiver pointer to the receiver
    registers interrupt for this receiver */
int kfedkit_register_interrupt (struct kfedkit_receiver_t *receiver)
{
    snprintf (receiver->device_name, FEDKIT_DEVNAME_SIZE, "fedkit_%d", receiver->index);
    receiver->device_name [FEDKIT_DEVNAME_SIZE - 1] = '\0';
    if (request_irq (receiver->device->irq, kfedkit_ISR, 
                     SA_SHIRQ, "fedkit_receiver" /* receiver->device_name*/, receiver)) {
        /* irq registering failed */
        return -1;
    } else {
        receiver->interrupt_registered = -1;
        return 0;
    }
}

void kfedkit_unregister_interrupt (struct kfedkit_receiver_t *receiver)
{
    if (receiver->interrupt_registered)
        free_irq (receiver->device->irq, receiver);
}

void kfedkit_unregister_receiver (struct kfedkit_board_t *board)
{
    struct kfedkit_receiver_t * receiver, **pp_receiver;
    int block_fifo_size;
    edprintf ("A\n");
    if (board->type != receiver_board) 
        printk (KERN_ERR "Something wrong going on in kfedkit_unregister_receiver: not a receiver\n");
    receiver = &board->receiver;
    /* lock */
    down (&receiver->lock);
    edprintf ("B\n");
    kfedkit_unregister_interrupt (receiver);
    /* deallocate stuff of the receiver */
    block_fifo_size = (receiver->block_number + 1) * sizeof (void *) + sizeof (struct _FK_data_fifo);
    edprintf ("C\n");
    if (receiver->block_fifo != NULL) {
        kfedkit_unnail_pages (receiver->block_fifo, block_fifo_size);
        edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->block_fifo);
	bigphysarea_free_pages (receiver->block_fifo);
    }
    edprintf ("D\n");
    if (receiver->blocks != NULL) {
        kfedkit_unnail_pages (receiver->blocks, receiver->block_size * receiver->block_number);
        edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->blocks);
	bigphysarea_free_pages (receiver->blocks);
    }
    edprintf ("E\n");
    if (receiver->wc_fifo != NULL) {
            kfedkit_unnail_pages (receiver->wc_fifo, sizeof (struct _FK_wc_fifo));
            edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->wc_fifo);
	    bigphysarea_free_pages (receiver->wc_fifo);
    }
    /* unmap the receiver's base addresses */
    edprintf ("F\n");
    if (receiver->map0 != NULL) {
        iounmap (receiver->map0);
    }
    /* remove the receiver from the receiver list */
    down (&kfedkit_receivers_lock);
    edprintf ("G\n");
    for (pp_receiver = &kfedkit_receivers;*pp_receiver != NULL;) {
        if (*pp_receiver == receiver) *pp_receiver = (*pp_receiver)->next;
        else pp_receiver = &(*pp_receiver)->next;
    }
    edprintf ("H\n");
    up (&kfedkit_receivers_lock);
    /* free the board structure */
    edprintf ("I\n");
    kfree (board);
    edprintf ("J\n");
}



/** fedkit_ioctl_receiver : various control specific to the receiver
        board(mainly allcate/delloc bigphys memory */
int kfedkit_ioctl_receiver (struct inode * inode_p, struct file *file_p, 
                   unsigned int cmd, unsigned long arg)
{
    struct kfedkit_board_t *board = file_p->private_data;
    struct kfedkit_receiver_t *receiver= NULL;
    if (board->type != receiver_board) return -EFAULT;
    receiver = &board->receiver;
    switch (cmd) {
    case FEDKIT_SETGET_RECEIVER_PARAMS:
    {
        struct _fedkit_receiver_parameters params;
        /* lock the file */
        down (&receiver->lock);
        /* if the memories were already allocated, this change is not allowed anymore */
        if (receiver->block_fifo != NULL) {
            up (&receiver->lock);
            return -EPERM;
        }
        /* get the parameters */
        if (copy_from_user(&params, (void *)arg, sizeof (struct _fedkit_receiver_parameters))) {
            up (&receiver->lock); 
            return -EFAULT;
        }
        /* Check the parameters and take note */
        if (params.block_number <=0 || params.block_size <= 0) {
            up (&receiver->lock);
            return -EINVAL;
        }
        receiver->block_number = params.block_number;
        receiver->block_size = params.block_size;
        receiver->do_allocate_blocks = params.do_allocate_blocks;
        /* no need to send back the parameters */
        up (&receiver->lock);
        return 0;
        break;
    }
    case FEDKIT_GET_RECEIVER_MEMBLOCKS:
    {
        struct _fedkit_memory_blocks mem_blocks;
        int block_fifo_size;
        /* lock the file */
        down (&receiver->lock);
        /* did we get the parameters yet ? */
        if (receiver->block_number == 0) {
            up (&receiver->lock);
            return -EPERM;
        }
        /* did we already allocate? */
        if (receiver->block_fifo != NULL) {
            up (&receiver->lock);
            return -EPERM;
        }
        /* allocate the bigphys buffers */
        block_fifo_size = (receiver->block_number + 1) * sizeof (void *) + sizeof (struct _FK_data_fifo);
        receiver->block_fifo = bigphysarea_alloc_pages ((block_fifo_size - 1) / PAGE_SIZE + 1, 0, GFP_KERNEL);
        receiver->wc_fifo = bigphysarea_alloc_pages ((sizeof (struct _FK_wc_fifo) - 1) / PAGE_SIZE + 1, 0, GFP_KERNEL);
        if (receiver->do_allocate_blocks) {
            receiver->blocks = (void*) bigphysarea_alloc_pages ((receiver->block_size * receiver->block_number - 1) / PAGE_SIZE + 1, 
                                                                   0, GFP_KERNEL);
#ifdef FEDKIT_PAINT_BIGPHYS
            {
                int i;
                uint32_t * block = (uint32_t *)receiver->blocks;
                for (i = 0; i < (receiver->block_size * receiver->block_number) ; i++) {
                    block[i] = virt_to_bus (&(block[i]));
                }
            }
#endif /* def FEDKIT_PAINT_BIGPHYS */
        } else {
            receiver->blocks = NULL;
        }
        /* check if anything went wrong */
        if (receiver->block_fifo == NULL || receiver->wc_fifo == NULL 
            || (receiver->do_allocate_blocks && receiver->blocks == NULL)) {
                if (receiver->block_fifo != NULL) { 
		    edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->block_fifo); 
		    bigphysarea_free_pages (receiver->block_fifo);
		}
                if (receiver->blocks != NULL) { 
		    edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->blocks); 
		    bigphysarea_free_pages (receiver->blocks);
		}
                if (receiver->wc_fifo != NULL) {
		    edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->wc_fifo); 
		    bigphysarea_free_pages (receiver->wc_fifo);
		}
                up(&receiver->lock);
                return -ENOMEM;
        }
        /* do the pages nailing */
        kfedkit_nail_pages (receiver->block_fifo, block_fifo_size);
        kfedkit_nail_pages (receiver->wc_fifo, sizeof (struct _FK_wc_fifo));
        if (receiver->do_allocate_blocks) 
            kfedkit_nail_pages (receiver->blocks, receiver->block_size * receiver->block_number);
        /* prepare the data_fifo structure */
        receiver->data_fifo = (struct _FK_data_fifo *) ((void *)receiver->block_fifo + 
                                                        (receiver->block_number + 1) * sizeof (void *));
        receiver->data_fifo_tab = receiver->block_fifo;
        receiver->data_fifo->read = 0;
        receiver->data_fifo->write = 0;
        receiver->data_fifo->done_read = 0;
        receiver->data_fifo->done_write = 0;
        receiver->data_fifo->size = (receiver->block_number + 1);
        /* buildup reply */
        mem_blocks.free_blocks_fifo_kernel = receiver->block_fifo;
        mem_blocks.free_blocks_fifo_physical = (void *)virt_to_bus (receiver->block_fifo);
        mem_blocks.wc_fifo_kernel = receiver->wc_fifo;
        mem_blocks.wc_fifo_physical = (void *)virt_to_bus (receiver->wc_fifo);
        if (receiver->do_allocate_blocks) {
            mem_blocks.data_blocks_kernel = receiver->blocks;
            mem_blocks.data_blocks_physical = (void *)virt_to_bus (receiver->blocks);
        } else {
            mem_blocks.data_blocks_kernel = NULL;
            mem_blocks.data_blocks_physical = NULL;
        }
        /* register the interrupt */
        if (kfedkit_register_interrupt (receiver)) {
            /* de allocate all the crap */
            if (receiver->block_fifo != NULL) {
                kfedkit_unnail_pages (receiver->block_fifo, block_fifo_size);
                edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->block_fifo);
		bigphysarea_free_pages (receiver->block_fifo);
            }
            if (receiver->blocks != NULL) {
                kfedkit_unnail_pages (receiver->blocks, receiver->block_size * receiver->block_number);
                edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->blocks);
		bigphysarea_free_pages (receiver->blocks);
            }
            if (receiver->wc_fifo != NULL) {
                kfedkit_unnail_pages (receiver->wc_fifo, sizeof (struct _FK_wc_fifo));
                edprintf ("About to free bigphys 0x%x\n", (unsigned int)receiver->wc_fifo);
		bigphysarea_free_pages (receiver->wc_fifo);
            }
            up (&receiver->lock); 
            return -EFAULT;
        }
        /* send reply */
        if (copy_to_user ((void *)arg, &mem_blocks, sizeof (struct _fedkit_memory_blocks))) {
            /* de allocate all the crap */
            if (receiver->block_fifo != NULL) {
                kfedkit_unnail_pages (receiver->block_fifo, block_fifo_size);
                bigphysarea_free_pages (receiver->block_fifo);
            }
            if (receiver->blocks != NULL) {
                kfedkit_unnail_pages (receiver->blocks, receiver->block_size * receiver->block_number);
                bigphysarea_free_pages (receiver->blocks);
            }
            if (receiver->wc_fifo != NULL) {
                kfedkit_unnail_pages (receiver->wc_fifo, sizeof (struct _FK_wc_fifo));
                bigphysarea_free_pages (receiver->wc_fifo);
            }
            kfedkit_unregister_interrupt (receiver);
            up (&receiver->lock); 
            return -EFAULT;            
        }
        /* complete */
        up (&receiver->lock);
        return 0;
        break;
    }
    default:
        return -EINVAL;
    }
    return -EFAULT;
}

/** fedkit_ioctl_receiver : various control specific to the receiver
        board(mainly allcate/delloc bigphys memory */
int kfedkit_ioctl_sender (struct inode * inode_p, struct file *file_p, 
                   unsigned int cmd, unsigned long arg)
{
#ifndef FEDKIT_NO_MASTER_INTS
    struct kfedkit_board_t *board = file_p->private_data;
    struct kfedkit_sender_t *sender = NULL;
    if (board->type != sender_board) return -EFAULT;
    sender = &board->sender;
    switch (cmd) {
    case FEDKIT_GET_SENDER_MEMBLOCKS:
    {
        /* The stuff fifo space is automatically allocated in the structure */
        struct _fedkit_sender_memory_blocks memblocks;        
        down (&sender->lock);
        memblocks.master_fifo_kernel = sender->master_fifo;
        memblocks.master_fifo_physical = (void*)virt_to_bus (memblocks.master_fifo_kernel);
        if (copy_to_user ((void*)arg, &memblocks, sizeof (struct _fedkit_sender_memory_blocks))) {
            printk (KERN_ERR "FEDKIT: failed to copy_to_user the sender memory block parameters\n");
            up (&sender->lock);
            return -EIO;
        }
        up (&sender->lock);
        return 0;
        break;
    }
    default:
        return -EINVAL;
    }
#endif /* ndef FEDKIT_NO_MASTER_INTS */
    return -EINVAL;
}


/** fedkit_ioctl : various controls for sender and reciever */
int kfedkit_ioctl (struct inode * inode_p, struct file *file_p, 
                   unsigned int cmd, unsigned long arg)
{
    struct kfedkit_board_t *board = file_p->private_data;
    switch (cmd) {
    case FEDKIT_GET_BOARD_ADDRESSES:
    {
        struct _fedkit_board_addresses reply = {0, 0, 0, 0};
        switch (board->type) {
        case sender_board:
            reply.base_address_0 = board->sender.base_address0;
            reply.base_address_1 = board->sender.base_address1;
            reply.base_address_2 = board->sender.base_address2;
            reply.FPGA_version = board->sender.FPGA_version;
            break;
        case receiver_board:
            reply.base_address_0 = board->receiver.base_address0;
            reply.base_address_1 = board->receiver.base_address1;
            reply.base_address_2 = board->receiver.base_address2;
            reply.FPGA_version = board->receiver.FPGA_version;
            break;
        default:
            return -EFAULT;
        }
        /* copy the reply to user sapce*/
        if (copy_to_user((void *)arg, &reply, sizeof (struct _fedkit_board_addresses)))
            return -EFAULT;
        return 0;
        break;
    }
    case FEDKIT_ALLOC_DBUFF:
    {
	return kfedkit_ioctl_alloc_dbuff(inode_p, file_p, cmd, arg);
	break;
    }
    case FEDKIT_FREE_DBUFF:
    {
	return kfedkit_ioctl_free_dbuff(inode_p, file_p, cmd, arg);
	break;
    }
	    
    default:
    {
        switch (board->type) {
        case sender_board:
            return kfedkit_ioctl_sender (inode_p, file_p, cmd, arg);
        case receiver_board:
            return kfedkit_ioctl_receiver (inode_p, file_p, cmd, arg);
        default:
            return -EFAULT;
        }
    }
    }
    return -EFAULT;
}

/** fedkit_mmap : has to handle both bigphys areas and DMA
        buffers. Should be fine with physical address*/
int kfedkit_mmap (struct file *file_p, struct vm_area_struct *vma)
{
    /* there could (should) be a different system call for mmaping the dbuffs, and the boards,
     * but the lack of checks make them identical */
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) 
    if (remap_pfn_range (vma, offset, vma->vm_pgoff, 
                         vma->vm_end - vma->vm_start, vma->vm_page_prot) ) {
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    if (remap_page_range (vma, vma->vm_start, offset, 
                          vma->vm_end - vma->vm_start, vma->vm_page_prot)) {
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,4,18) && RED_HAT_LINUX_KERNEL
    if (remap_page_range (vma, vma->vm_start, offset, 
                          vma->vm_end - vma->vm_start, vma->vm_page_prot)) {
#else
    if (remap_page_range (vma->vm_start, offset, 
                              vma->vm_end - vma->vm_start, vma->vm_page_prot)) {
#endif
        eprintf ("kfedkit_mmap: fail to remap start=0x%08lx, end=0x%08lx, ofset = 0x%08lx\n", 
                 vma->vm_start, vma->vm_end, offset);
        return -EAGAIN;
    }
    idprintf ("In kfedkir_mmap : mapped successfully vma with :\n");
    idprintf ("*start  = 0x%08x\n", (int)vma->vm_start);
    idprintf ("*offset = 0x%08x\n", (int)offset);
    idprintf ("*end =    0x%08x\n", (int)vma->vm_end);
    idprintf ("*(size =  0x%08x)\n",(int)(vma->vm_end-vma->vm_start));
    return (0);   /* happy */
}

/** fedkit release : close the file and deallocate everything */
int kfedkit_release (struct inode *inode_p, struct file *file_p)
{
    struct kfedkit_board_t *board = file_p->private_data;
    edprintf ("kfedkit_release: board->type = %d\n", (int)board->type);
    /* Free all the dbuffs attached to the board. This is safe as the kernel only
     * calls the release function on a file desciptor when the last of open file descriptor
     * AND memory maps for it is closed. All the dbuffs are hence not mapped anymore */
    kfedkit_dbuffs_free (board->dbuffs);
    switch (board->type) {
    case receiver_board:
    {
        edprintf ("kfedkit_release: receiver_board\n");
        kfedkit_unregister_receiver (board);
        break;
    }
    case sender_board:
    {
        edprintf ("kfedkit_release: sender_board\n");
        kfedkit_unregister_sender (board);
        break;
    }
    default:
        edprintf ("kfedkit_release: default\n");
        return -EFAULT;
    };
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_DEC_USE_COUNT;
#endif
    return 0;
}


MODULE_LICENSE("GPL");
